/* Copyright (c) 2004 Limor Fried <ladyada@media.mit.edu>, MIT Media Lab

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <inttypes.h>
#include "uipopt.h"

#define ETHERNET_HEADER_LENGTH 0x0E

/*****************************************************************************
*
* MAC address assigned to the CS8900a
*
*****************************************************************************/
#define MYMAC_0 UIP_ETHADDR0
#define MYMAC_1 UIP_ETHADDR1
#define MYMAC_2 UIP_ETHADDR2
#define MYMAC_3 UIP_ETHADDR3
#define MYMAC_4 UIP_ETHADDR4
#define MYMAC_5 UIP_ETHADDR5


#define NOP asm("nop")

#define DEBUGGING 1

#define DEBUG(x) if (DEBUGGING) x

#define HI(x) (x>>8)&0xFF
#define LO(x) x&0xFF

#define CRYSTAL_AEN    PD2
#define CRYSTAL_IOW    PD3
#define CRYSTAL_IOR    PD4
#define CRYSTAL_RESET  PD5

#define DATAPORT  PORTC
#define DATAPORTTRIS DDRC
#define CONTROLPORT PORTD
#define ADDRPORT PORTB

typedef unsigned char byte;
typedef unsigned char uchar;
typedef unsigned int uint;

// the uIP interface functions;
void cs8900aInit(void);
uint cs8900aBeginPacketRetreive(void); // returns packet length
void cs8900aRetreivePacketData(uchar *localBuffer, uint length);
void cs8900aEndPacketRetreive(void);
void cs8900aBeginPacketSend(uint packetLength);
void cs8900aSendPacketData(uchar * localBuffer, uint length);
void cs8900aEndPacketSend(void);


/********** Functions **********/
// low level initialization functions
void crystalEnable(void);
byte crystalVerify(void);
void crystalReset(void);
void crystalInit(byte promisc, byte *MAC);

// low level IO functions
void crystalSetAddr(byte addr);
void crystalSetPointer(uint16_t addr);
void crystalIOWrite(byte addr, byte data);
byte crystalIORead(byte addr);
byte crystalPoll(void);

uint16_t crystalReadData(uint16_t addr);
void crystalWriteData(uint16_t addr, uint16_t data);

uint16_t crystalRxWord(void);

void crystalDropFrame(void);
byte crystalRxBytes(byte *buff, int num);
/********* Defines ************/
#define RX_TX_DATA      0x00
#define RX_TX_DATA1     0x02
#define TX_CMD          0x04
#define TX_LEN          0x06
#define ISQ             0x08
#define PORTPTR         0x0a
#define PORTDATA        0x0c
#define PORTDATA1       0x0e

#define SELF_CTL_RESET     0x0040

#define EISA      	0x0000
#define PRODID    	0x0002

#define RX_CFG			0x0102
#define RX_CTL			0x0104
#define TX_CFG			0x0106

#define	LINE_CTL		0x0112
#define SELF_CTL  		0x0114
#define BUS_CTL			0x0116
#define TEST_CTL		0x0118
#define  RX_EVT      0x0124
#define  TX_EVT      0x0128


#define SELF_STATUS 0x0136
#define BUS_STATUS 0x0138
#define IND_ADDR		0x0158         // Individual addr (MAC)
#define TX_BID_ERR   0x0080

// flags
#define TX_CMD_START_5         0x0000
#define TX_CMD_START_381       0x0080
#define TX_CMD_START_1021      0x0040
#define TX_CMD_START_ALL       0x00C0
#define TX_CMD_FORCE           0x0100
#define TX_CMD_ONE_COLL        0x0200
#define TX_CMD_NO_CRC          0x1000
#define TX_CMD_NO_PAD          0x2000

#define BUS_ST_TX_BID_ERR        0x0080
#define BUS_ST_RDY4TXNOW         0x0100

#define TEST_CTL_DIS_LT    		0x0080
#define TEST_CTL_ENDEC_LP  		0x0200
#define TEST_CTL_AUI_LOOP  		0x0400
#define TEST_CTL_DIS_BKOFF 		0x0800
#define TEST_CTL_FDX       		0x4000

#define LINE_CTL_RX_ON         0x0040
#define LINE_CTL_TX_ON         0x0080
#define LINE_CTL_AUI_ONLY      0x0100
#define LINE_CTL_10BASET       0x0000

#define RX_EVENT_RX_OK         0x0100
#define RX_EVENT_IND_ADDR      0x0400
#define RX_EVENT_BCAST         0x0800
#define RX_EVENT_CRC_ERR       0x1000
#define RX_EVENT_RUNT          0x2000
#define RX_EVENT_X_DATA        0x4000

#define RX_CTL_PROMISC           0x0080
#define RX_CTL_RX_OK_A           0x0100
#define RX_CTL_MCAST_A           0x0200
#define RX_CTL_IND_A             0x0400
#define RX_CTL_BCAST_A           0x0800
#define RX_CTL_CRC_ERR_A         0x1000
#define RX_CTL_RUNT_A            0x2000
#define RX_CTL_X_DATA_A          0x4000

#define RX_CFG_SKIP            0x0040
#define RX_CFG_RX_OK_IE        0x0100
#define RX_CFG_CRC_ERR_IE      0x1000
#define RX_CFG_RUNT_IE         0x2000
#define RX_CFG_X_DATA_IE       0x4000
