#include <inttypes.h>

#define DHCP_CLIENT_PORT 68
#define DHCP_SERVER_PORT 67

//message types
#define DHCPDISCOVER 1
#define DHCPOFFER 2
#define DHCPREQUEST 3
#define DHCPDECLINE 4
#define DHCPACK 5
#define DHCPNAK 6
#define DHCPRELEASE 7
#define DHCPINFORM 8

//op codes
#define DHCP_BOOTREQUEST 1
#define DHCP_BOOTREPLY 2

#define DHCP_OPT_ROUTER 0x03
#define DHCP_OPT_HOSTNAME 0x0c
#define DHCP_OPT_REQIP 0x32
#define DHCP_OPT_DNS 0x06
#define DHCP_OPT_PARAMREQ 0x37
#define DHCP_OPT_VENDORID 0x3c
#define DHCP_OPT_CLIENTID 0x3d
#define DHCP_OPT_NETMASK 0x01
#define DHCP_OPT_REQUESTED_IP 50
#define DHCP_OPT_LEASE_TIME 51
#define DHCP_OPT_MESSAGE_TYPE 53
#define DHCP_OPT_T1_TIME 58
#define DHCP_OPT_T2_TIME 59

#define DHCP_PARAMREQ_SUBNETMASK  0x01;
#define DHCP_PARAMREQ_DOMAINNAME  0x0f;
#define DHCP_PARAMREQ_ROUTER      0x03;
#define DHCP_PARAMREQ_DNS         0x06;
#define DHCP_PARAMREQ_ROUTERDISC  0x1f;
#define DHCP_PARAMREQ_STATICROUTE 0x21;
#define DHCP_ENDOPT 0xFF;

//dhcp "magic cookie": first 4 bytes in options section of BOOTP message
// 99 130 83 99
#define DHCP_OPTION_COOKIE  0x63825363

#define DHCP_OPTIONSIZE 128
typedef enum {DHCP_SUCCESS, DHCP_ERROR} dhcpErrorCode;

typedef enum {DHCP_BEGIN, DHCP_DISCOVER, DHCP_REQUEST, DHCP_DONE} dhcpStatus;

/*
   0                   1                   2                   3
   0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |     op (1)    |   htype (1)   |   hlen (1)    |   hops (1)    |
   +---------------+---------------+---------------+---------------+
   |                            xid (4)                            |
   +-------------------------------+-------------------------------+
   |           secs (2)            |           flags (2)           |
   +-------------------------------+-------------------------------+
   |                          ciaddr  (4)                          |
   +---------------------------------------------------------------+
   |                          yiaddr  (4)                          |
   +---------------------------------------------------------------+
   |                          siaddr  (4)                          |
   +---------------------------------------------------------------+
   |                          giaddr  (4)                          |
   +---------------------------------------------------------------+
   |                                                               |
   |                          chaddr  (16)                         |
   |                                                               |
   |                                                               |
   +---------------------------------------------------------------+
   |                                                               |
   |                          sname   (64)                         |
   +---------------------------------------------------------------+
   |                                                               |
   |                          file    (128)                        |
   +---------------------------------------------------------------+
   |                                                               |
   |                          options (variable)                   |
   +---------------------------------------------------------------+
*/

typedef struct dhcpPacket{
  uint8_t op;             //Operation
  uint8_t htype;          //Hardware Type
  uint8_t hlen;           //Hardware Address Length
  uint8_t hops;           //Hops
  uint16_t xid[2];         //Transaction ID
  uint16_t secs;        //Seconds
  uint16_t flags;       //Flags
  uint16_t ciaddr[2];      //Client IP Address
  uint16_t yiaddr[2];      //Your IP Address
  uint16_t siaddr[2];      //Server IP Address
  uint16_t giaddr[2];      //Gateway IP Address
  uint8_t chaddr[16];     //Client Hardware Address
  uint8_t sname[64];      //Server Name
  uint8_t file[128];       //File Name
  uint8_t options[DHCP_OPTIONSIZE];   //Options (up to 312 bytes)
} uip_dhcp_hdr;


#define DHCP_BASE_LEN 236 // w/o options

#define DHCPBUF  ((uip_dhcp_hdr *)&uip_buf[UIP_LLH_LEN + UDP_HEADER_LEN + IP_HEADER_LEN])


void dhcp_init(void);
dhcpErrorCode dhcp_send_discover(uint16_t xid1, uint16_t xid2);
void dhcp_send_packet(void);
void dhcp_listen(void);
char dhcp_messagetype(void);
void dhcp_recv_offer(void);
uint8_t *dhcp_findoption(uint8_t optionname);
void dhcp_send_request(void );
