/* Copyright (c) 2004 Limor Fried <ladyada@media.mit.edu>, MIT Media Lab

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/


#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include "main.h"
#include "uip.h"
#include "udp.h"
#include "uip_arp.h"
#include "nic.h"
#include "dnslookup.h"

uint16_t dns_nameserver[2];
char xactionid;

void dns_set_nameserver(uint16_t *ns) {
  dns_nameserver[0] = ns[0];
  dns_nameserver[1] = ns[1];
}

void dns_init(uint16_t *ns) {
  dns_set_nameserver(ns);
  xactionid = 0;
}

uint16_t dns_resolved[2];
// request null-terminated name
uint8_t lookup(char *name) {
  uint8_t i;
  uint8_t c;
  uint16_t word;
  char namelen;
  uint8_t *dnsdata;

  unsigned char timeout;
  int packetlen = 0;

  xactionid++;

  namelen = strlen(name);
  memset(DNSBUF, 0, sizeof(struct uip_dns));
  DNSBUF->transaction_id = xactionid;
  DNSBUF->flags = HTONS(0x0100);  // recursive query
  DNSBUF->questions = HTONS(1);
  DNSBUF->answer_RRs = 0;
  DNSBUF->auth_RRs = 0;
  DNSBUF->add_RRs = 0;

  c = 0;
  strcpy(&(DNSBUF->data[1]), name);

  for (i = 0; i<namelen+1; i++) {
    if (name[i] == '.' || name[i] == 0) {
      DNSBUF->data[i-c] = c;
      c = 0;
    } else {
      DNSBUF->data[i+1] = name[i];
      c++;
    }
  }

  DNSBUF->data[namelen+3] = 1;
  DNSBUF->data[namelen+5] = 1;
  DEBUG = 0;
  udp_send_packet(uip_hostaddr, dns_nameserver, 
		  DNS_CLIENT_PORT, DNS_SERVER_PORT,
		  12+namelen+6);

  DEBUG = 0;

  // ok now listen for response
 
  timeout = getTimer0();
  while (getTimer0() - timeout < 3) {       // ~3 second timeout
    do {
      packetlen = nic_poll();
    } while (packetlen == 0);

    // find a TCP UDP packet
    if ( (ETHBUF->type == htons(UIP_ETHTYPE_IP)) &&
	 (UDPBUF->vhl == 0x45) &&
	 (UDPBUF->proto == UIP_PROTO_UDP) &&
	 (UDPBUF->srcport == HTONS(DNS_SERVER_PORT)) && 
	 (UDPBUF->destport == HTONS(DNS_CLIENT_PORT)))
      {
	if (DNSBUF->transaction_id != xactionid) { // must be our transaction
	  continue;
	}
	if (!(NTOHS(DNSBUF->flags) & 0x8000)) { // must be a response
	  continue;
	}
	if ((NTOHS(DNSBUF->flags) & 0xF) == 0x3) { // stop if theres no answer
	  return DNS_NOSUCHNAME;
	}
	/*
	printf("qs: %d ans: %d auth: %d add: %d\n\r", 
	       NTOHS(DNSBUF->questions),
	       NTOHS(DNSBUF->answer_RRs),
	       NTOHS(DNSBUF->auth_RRs),
	       NTOHS(DNSBUF->add_RRs));
	*/
	if ((DNSBUF->answer_RRs == 0x0) && (DNSBUF->auth_RRs == 0x0)) {
	  continue;
	}  // if theres no data in it, ignore?

	dnsdata = DNSBUF->data;  // make a pointer to the data portion of the packet
	
	// go through the questions
	for (i=0; i < NTOHS(DNSBUF->questions); i++) {
	  while (1) {
	    c = dnsdata[0];
	    dnsdata++;
	    if (c == 0) {
	      dnsdata += 4;
	      break;
	    }
	    if ((c & 0xC0) == 0xC0) {
	      dnsdata += 1;
	      break;
	    }
	    dnsdata += c;
	  }
	}
	
	// ok at the answers or auth point
	for (i=0; i < NTOHS(DNSBUF->answer_RRs) + NTOHS(DNSBUF->auth_RRs); i++) {
	  // get past the name
	  dnsdata = inc_past_name(dnsdata);

	  word = dnsdata[0] << 8 | dnsdata[1];
	  //printf("type %x ", word);
	  if (word != 0x0001) {  // must be host addr
	    dnsdata += 8;
	    word = dnsdata[0] << 8 | dnsdata[1];
	    //printf("record len %d\n\r", word);
	    dnsdata += word +2;
	    continue;
	  }
	  dnsdata+=2;

	  word = dnsdata[0] << 8 | dnsdata[1];
	  //printf("class %x ", word);
	  if (word != 0x0001) {  // must be inet
	    dnsdata += 6;
	    word = dnsdata[0] << 8 | dnsdata[1];
	    dnsdata += word + 2;
	    continue;
	  }
	  dnsdata += 2;
	  
	  dnsdata += 4;   // time to live

	  dnsdata += 2;   // should be 0x0004 (addr length)
	  dns_resolved[0] = dnsdata[1] << 8 | dnsdata[0];
	  dns_resolved[1] = dnsdata[3] << 8 | dnsdata[2];
	  return DNS_OK;
	}
      }
  }    
  return DNS_TIMEOUT;
}

// a stupid helper. does the work of incrementing the buffer past the
// dns name, handles compression.
uint8_t *inc_past_name(uint8_t *buff) {
  uint8_t c;
  while (1) {
    c = buff[0];
    buff++;
    if (c == 0) {
      break;
    }
    if ((c & 0xC0) == 0xC0) { // this is a pointer
      buff++;
      break;
    }
    //uart_putchar('0'+c);
    buff += c;
  }
  return buff;
}
