/* Copyright (c) 2004 Limor Fried <ladyada@media.mit.edu>, MIT Media Lab

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <inttypes.h>
#include <string.h>
#include <stdio.h>
#include "tcp.h"
#include "uip.h"
#include "main.h"
#include "uip_arch.h"
#include "uip_arp.h"
#include "nic.h"

tcp_conn client_conns[MAX_CLIENTCONN];
uint16_t nextclientport;

void tcp_init(void) {
  int i;
  tcp_conn *conn;

  for (i=0; i<MAX_CLIENTCONN; i++) {
    conn = &(client_conns[i]);
    memset(conn, 0, sizeof(tcp_conn));
    conn->state = TCP_STATE_CLOSED;
  }

  nextclientport = rnd();
}

void nullhandler(tcp_conn *conn) {
}

uint32_t bignum;
void syn_recv_handler(tcp_conn *conn) {
  
  bignum = TCPBUF->ackno[0]; bignum <<= 8;
  bignum |= TCPBUF->ackno[1]; bignum <<= 8;
  bignum |= TCPBUF->ackno[2]; bignum <<= 8;
  bignum |= TCPBUF->ackno[3];

  if ((TCPBUF->flags & TCP_FLAGS_SYN) && (bignum == conn->seq_num)) {
    conn->state = TCP_STATE_ESTABLISHED;
    putstring("Connected to "); printIP(conn->destaddr); putstring(".\n\r");
  }
}

void syn_sent_handler(tcp_conn *conn) {
  bignum = TCPBUF->ackno[0]; bignum <<= 8;
  bignum |= TCPBUF->ackno[1]; bignum <<= 8;
  bignum |= TCPBUF->ackno[2]; bignum <<= 8;
  bignum |= TCPBUF->ackno[3];
  
  if ((TCPBUF->flags & TCP_FLAGS_SYN)) {
    conn->ack_num = TCPBUF->seqno[0]; conn->ack_num <<= 8;
    conn->ack_num |= TCPBUF->seqno[1]; conn->ack_num <<= 8;
    conn->ack_num |= TCPBUF->seqno[2]; conn->ack_num <<= 8;
    conn->ack_num |= TCPBUF->seqno[3]; 

    conn->ack_num++;
    tcp_send_packet(conn, 0, TCP_FLAGS_ACK, 0);
    if ((TCPBUF->flags & TCP_FLAGS_ACK) && (bignum == conn->seq_num)) {
      conn->state = TCP_STATE_ESTABLISHED;
      putstring("Connected to "); printIP(conn->destaddr); putstring(".\n\r");
      return conn->establishedcallback(conn, 0, 0); // theres no data in establishment
    } else {
      conn->state = TCP_STATE_SYN_RECV;
      putstring("hmm, just SYN?");
      conn->callback = &syn_recv_handler;
    }
    //send final ack
  } else {
    // ???
    printf("got strange response, flags = %x\n\r", TCPBUF->flags);
  }
}

void tcp_process() {
  int i;
  uint16_t len;
  tcp_conn *conn;

  // find the connection that matches (same dest/src ip/port)
  for (i=0; i<MAX_CLIENTCONN; i++) {
    conn = &(client_conns[i]);


    if ( (TCPBUF->srcipaddr[0] = conn->destaddr[0]) &&
	 (TCPBUF->srcipaddr[1] = conn->destaddr[1]) &&
	 (TCPBUF->destipaddr[0] = conn->srcaddr[0]) &&
	 (TCPBUF->destipaddr[1] = conn->srcaddr[1]) &&
	 (TCPBUF->srcport = HTONS(conn->srcport)) &&
	 (TCPBUF->destport = HTONS(conn->destport))) {

      //putstring("found a match!\n\r");
      //uart_putchar('*');
      len = TCPBUF->len[0]; len <<= 8;
      len |= TCPBUF->len[1];
      len -= 20 + (TCPBUF->tcpoffset >> 2);
      
      bignum = TCPBUF->seqno[0]; bignum <<= 8;
      bignum |= TCPBUF->seqno[1]; bignum <<= 8;
      bignum |= TCPBUF->seqno[2]; bignum <<= 8;
      bignum |= TCPBUF->seqno[3];

      //printf("flags = 0x%x len = %d\n\r", TCPBUF->flags, len);
      if (TCPBUF->flags & TCP_FLAGS_RST) {
	conn->state = TCP_STATE_CLOSED;
	conn->callback = &nullhandler;
	putstring("Connection reset\n\r");
      } else if (TCPBUF->flags & TCP_FLAGS_FIN) {
	// handle this better?
	conn->state = TCP_STATE_CLOSED;
	conn->callback = &nullhandler;


	//printf("Seq #%ld vs ack #%ld (%d)\n\r", bignum, conn->ack_num, len);
	
	if (bignum < conn->ack_num) {  // resend of old data....
	  // resend the old ack!
	  tcp_send_packet(conn, 0, TCP_FLAGS_ACK | TCP_FLAGS_FIN, 0);
	} else {
	  conn->ack_num+=len;
	  conn->establishedcallback(conn, 
				    (uint8_t *)TCPBUF+20+(TCPBUF->tcpoffset >> 2), 
				    len);
	  
	  
	  
	  tcp_send_packet(conn, 0, TCP_FLAGS_ACK | TCP_FLAGS_FIN, 0);
	  putstring("\n\rConnection closed\n\r");
	}
	
      }  else if (conn->state == TCP_STATE_ESTABLISHED) {
	// received "normal data", send it to the callback for this connection

	if (len != 0) {
	  //printf("Seq #%ld vs ack #%ld (%d)\n\r", bignum, conn->ack_num, len);
	}

	if (bignum < conn->ack_num) {  // resend of old data....
	  // resend the old ack!
	  tcp_send_packet(conn, 0, TCP_FLAGS_ACK, 0);
	} else if (len != 0) {
	  conn->establishedcallback(conn, 
				    (uint8_t *)TCPBUF+20+(TCPBUF->tcpoffset >> 2), 
				    len);
	  // ack it!
	  conn->ack_num += len;
	  tcp_send_packet(conn, 0, TCP_FLAGS_ACK, 0);
	}
      } else {
	conn->callback(conn);
      }
    }
  }
}


uint8_t tcp_open(uint16_t *destaddr, uint16_t destport,
		 void (*establishedfunction) (tcp_conn *, uint8_t *, uint16_t)) {
  int i;
  tcp_conn *conn;

  // find an open client connection
  for (i = 0; i<MAX_CLIENTCONN; i++) {
     conn = &(client_conns[i]);
     if (conn->state == TCP_STATE_CLOSED)
       break;
  }
  if (i == MAX_CLIENTCONN) 
    return TCP_NOFREECONNECTIONS;

  conn->srcaddr[0] = uip_hostaddr[0];
  conn->srcaddr[1] = uip_hostaddr[1];
  conn->destaddr[0] = destaddr[0];
  conn->destaddr[1] = destaddr[1];
  conn->srcport = nextclientport++; 
  conn->destport = destport;

  conn->seq_num = rnd();
  conn->seq_num <<= 16;
  conn->seq_num |= rnd();

  // set MSS to 250 bytes of data (so 300 bytes max)
  uip_appdata = (u8_t *)TCPBUF + 40;
  uip_appdata[0] = 0x2;
  uip_appdata[1] = 0x4;
  uip_appdata[2] = UIP_TCP_MSS>>8;
  uip_appdata[3] = UIP_TCP_MSS & 0xff;

  // send syn
  putstring("Trying "); printIP(destaddr); putstring("...\n\r");
  tcp_send_packet(conn, 0, TCP_FLAGS_SYN, 4); // 4 bytes options
  conn->state = TCP_STATE_SYN_SENT;
  conn->seq_num++;

  conn->callback = &syn_sent_handler;
  conn->establishedcallback = establishedfunction;
  return 0;
}

void tcp_send_packet(tcp_conn *conn, uint16_t datalen, 
		     uint8_t flags, uint8_t optlen) {
  uint16_t totallen;
  totallen = datalen + UIP_TCPIP_HLEN + optlen;

  TCPBUF->vhl = 0x45;
  TCPBUF->tos = 0x0;
  TCPBUF->len[0] = totallen >> 8;
  TCPBUF->len[1] = totallen;
  TCPBUF->ipid[0] = 0;
  TCPBUF->ipid[1] = 0;
  TCPBUF->ipoffset[0] = 0x40;    // flags = don't fragment
  TCPBUF->ipoffset[1] = 0x0;
  TCPBUF->ttl = 128;
  TCPBUF->proto = UIP_PROTO_TCP;
  TCPBUF->srcipaddr[0] = conn->srcaddr[0];
  TCPBUF->srcipaddr[1] = conn->srcaddr[1];
  TCPBUF->destipaddr[0] = conn->destaddr[0];
  TCPBUF->destipaddr[1] = conn->destaddr[1];
  TCPBUF->ipchksum = 0;
  TCPBUF->ipchksum = ~(uip_ipchksum());
  
  TCPBUF->srcport = HTONS(conn->srcport);
  TCPBUF->destport = HTONS(conn->destport);
  TCPBUF->seqno[0] = conn->seq_num >> 24;
  TCPBUF->seqno[1] = conn->seq_num >> 16;
  TCPBUF->seqno[2] = conn->seq_num >> 8;
  TCPBUF->seqno[3] = conn->seq_num;
  TCPBUF->ackno[0] = conn->ack_num >> 24;
  TCPBUF->ackno[1] = conn->ack_num >> 16;
  TCPBUF->ackno[2] = conn->ack_num >> 8;
  TCPBUF->ackno[3] = conn->ack_num;
  TCPBUF->tcpoffset = (20 + optlen)<<2 ;
  TCPBUF->flags = flags;
  TCPBUF->wnd[0] = UIP_TCP_MSS >> 8; 
  TCPBUF->wnd[1] =  UIP_TCP_MSS;
  
  TCPBUF->tcpchksum = 0;
  //TCPBUF->tcpchksum = ~tcp_checksum(totallen);
  uip_appdata = (u8_t *)TCPBUF + 40;
  TCPBUF->tcpchksum = ~uip_tcpchksum();

  uip_len = totallen;
  uip_arp_out();
  NICBeginPacketSend(uip_len);

  NICSendPacketData(uip_buf, uip_len);
  NICEndPacketSend();

  
}


uint8_t tcp_send(tcp_conn *conn, uint8_t *data, uint16_t len, 
	      void (*establishedfunction) (tcp_conn *, uint8_t *, uint16_t)) {
  unsigned char timeout;
  uint16_t packetlen = 0;

  tcp_send_packet(conn, len, TCP_FLAGS_ACK | TCP_FLAGS_PSH, 0);
  // wait for ack

  timeout = getTimer0();
  while (getTimer0() == timeout) {       // ~1 second timeout
    do {
      packetlen = nic_poll();
    } while (packetlen == 0);

    // find a TCP IP packet
    if ( (ETHBUF->type == htons(UIP_ETHTYPE_IP)) &&
	 (TCPBUF->vhl == 0x45) &&
	 (TCPBUF->proto == UIP_PROTO_TCP) &&
	 (TCPBUF->srcipaddr[0] = conn->destaddr[0]) &&
	 (TCPBUF->srcipaddr[1] = conn->destaddr[1]) &&
	 (TCPBUF->destipaddr[0] = conn->srcaddr[0]) &&
	 (TCPBUF->destipaddr[1] = conn->srcaddr[1]) &&
	 (TCPBUF->srcport = HTONS(conn->srcport)) &&
	 (TCPBUF->destport = HTONS(conn->destport)) &&
	 (TCPBUF->flags & TCP_FLAGS_ACK))
      {
	bignum = TCPBUF->ackno[0]; bignum <<= 8;
	bignum |= TCPBUF->ackno[1]; bignum <<= 8;
	bignum |= TCPBUF->ackno[2]; bignum <<= 8;
	bignum |= TCPBUF->ackno[3];
	if (bignum != (conn->seq_num + len)) {
	  printf("? %ld vs %ld\n\r", bignum, (conn->seq_num + len));
	}
	return TCP_OK;
      }
  }
  return TCP_TIMEDOUT;
}

uint16_t tcp_checksum(uint16_t len) {
  uint8_t padding = 0;
  uint8_t *buff = (uint8_t *)TCPBUF+40;
  uint16_t w;
  uint32_t sum;
    uint16_t i;

  if (len & 1) {
    padding++;
    buff[len] = 0;
  }
  
  sum = 0;
  for (i = 0; i< len+padding; i+=2) {
    w = buff[i];
    w <<= 8;
    w |= buff[i+1];
    sum += w;
  }
  
  sum += TCPBUF->srcipaddr[0];
  sum += TCPBUF->srcipaddr[1];
  sum += TCPBUF->destipaddr[0];
  sum += TCPBUF->destipaddr[1];

  sum += 0x06 + len;

  while (sum >> 16)
    sum = (sum & 0xFFFF) + (sum >> 16);
  
  return sum;
}


