/* Copyright (c) 2004 Limor Fried <ladyada@media.mit.edu>, MIT Media Lab

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef _TCP_H_
#define _TCP_H_

#define MAX_CLIENTCONN 3
#define TCP_NOFREECONNECTIONS 1
#define TCP_TIMEDOUT 2
#define TCP_OK 4

#define TCP_FLAGS_FIN 0x01
#define TCP_FLAGS_SYN 0x02
#define TCP_FLAGS_RST 0x04
#define TCP_FLAGS_PSH 0x08
#define TCP_FLAGS_ACK 0x10



#define TCP_STATE_NONE 0
#define TCP_STATE_ESTABLISHED 1
#define TCP_STATE_SYN_SENT 2
#define TCP_STATE_SYN_RECV 3
#define TCP_STATE_FIN_WAIT 4
#define TCP_STATE_TIME_WAIT 5
#define TCP_STATE_CLOSED 6
#define TCP_STATE_CLOSE_WAIT 7
#define TCP_STATE_LAST_ACK 8
#define TCP_STATE_LISTEN 9

typedef struct tcp_conn_struct {

  uint16_t srcaddr[2];
  uint16_t destaddr[2];
  uint16_t srcport, destport;

  uint32_t seq_num, ack_num;

  uint8_t state;

  void (*callback) (struct tcp_conn_struct *conn);
  void (*establishedcallback) (struct tcp_conn_struct *, uint8_t *, uint16_t);
} tcp_conn;

void tcp_init(void);

uint8_t tcp_open(uint16_t *destaddr, uint16_t destport,
		 void (*establishedfunction) (tcp_conn *, uint8_t *, uint16_t));
uint8_t tcp_send(tcp_conn *conn, uint8_t *data, uint16_t len, 
		 void (*establishedfunction) (tcp_conn *, uint8_t *, uint16_t));
uint16_t tcp_checksum(uint16_t len);
void tcp_send_packet(tcp_conn *connection, uint16_t datalen, 
		     uint8_t flags, uint8_t optlen);
void syn_sent_handler(tcp_conn *conn);
void tcp_process(void);

#endif // _TCP_H_
